//#define EASY
static unsigned over_lines=0;

// Standard C++ includes:
#include <cstdio>    // for std::puts
#include <cstdlib>   // for std::rand
#include <cstring>   // for std::memcpy
#include <algorithm> // for std::abs, std::max, std::random_shuffle
#include <utility>   // for std::pair
#include <vector>    // for std::vector

// DJGPP-specific includes:
#include <conio.h>      // for kbhit,getch
#include <sys/farptr.h> // for _farpeekl, _farpokew
#include <go32.h>       // for _dos_ds
#include <dpmi.h>       // for __dpmi_yield
#include <dos.h>

// Coroutine library. With thanks to Simon Tatham.
#define ccrVars      struct ccrR { int p; ccrR() : p(0) { } } ccrL
#define ccrBegin(x)  auto& ccrC=(x); switch(ccrC.ccrL.p) { case 0:;
#define ccrReturn(z) do { ccrC.ccrL.p=__LINE__; return(z); case __LINE__:; } while(0)
#define ccrFinish(z) } ccrC.ccrL.p=0; return(z)

// Field & rendering definitions
const auto Width = 12, Height = 25; // Width and height, including borders.
const auto VidWidth = 40u;
const auto Occ = 0x10000u;
const auto Empty = 0x12Eu, Border = Occ+0x7DBu, Garbage = Occ+0x6DBu;

// In order to access BIOS timer and VGA memory:
#define Timer _farpeekl(_dos_ds, 0x46C)
struct { struct Proxy { unsigned offset;
    void operator=(unsigned short value) const { _farpokew(_dos_ds, offset, value); } };
    Proxy operator[] (unsigned offset) const { return Proxy( {0xB8000+2u*offset} ); }
} const VidMem = {}; // Simulate an array that is at B800:0000

unsigned long SoundOffTimer = 0, SoundFreq = 0, SoundModuloDec = 0, SoundModuloTimer = 0;
static void SoundSet(unsigned freq)
{
    unsigned long scale = 0x1234DDul / freq;
    outportb(0x42, scale & 0xFF);
    outportb(0x42, scale >> 8);
}
static void Sound(unsigned freq, unsigned duration)
{
    outportb(0x43, 0xB6);
    SoundSet(SoundFreq = freq);
    outportb(0x61, inportb(0x61) | 3);
    SoundOffTimer = Timer + duration;
    SoundModuloDec = 0;
}
static void UpdateSound()
{
    if(Timer > SoundOffTimer)
        { outportb(0x61, inportb(0x61) & ~3);
          SoundOffTimer = Timer+1000000;
          SoundFreq = 0; }
    else if(SoundFreq > 0 && SoundFreq <= 50)
        SoundSet( 5 + std::rand() % (40-SoundModuloDec) );
    if(SoundModuloTimer != Timer)
    {
        SoundModuloTimer = Timer;
        if(SoundModuloDec < 35)
            SoundModuloDec += 1;
    }
}

// Game engine
struct Piece
{
    struct { unsigned short s[4]; } shape;
    int x:8, y:8, color:14;
    unsigned r:2;

    template<typename T> // walkthrough operator
    inline bool operator>(T it) const
    {
        for(int q=shape.s[r],p=1,by=0; by<4; ++by)
            for(int bx=0; bx<4; ++bx, p<<=1)
                if((q&p) && it(x+bx, y+by)) return true;
        return false;
    }
    template<typename T> // transmogrify operator
    inline Piece operator*(T it) const { Piece res(*this); it(res); return res; }
};

template<bool DoDraw>
struct TetrisArea
{
    int Area[Height][Width];
    unsigned RenderX;
    unsigned n_full, list_full, animx;
    unsigned long timer;
    struct { ccrVars; } cascadescope;
public:
    TetrisArea(unsigned x=0) : RenderX(x) { }

    bool Occupied(int x,int y) const
        { return x<1 || (x>Width-2) || (y>=0 && (Area[y][x] & Occ)); }

    template<typename T>
    inline void DrawRow(unsigned y, T get)
    {
        for(int x=1; x<Width-1; ++x) DrawBlock(x,y, get(x));
    }

    inline bool TestFully(unsigned y, bool state) const
    {
        for(int x=1; x<Width-1; ++x) if(state != !!(Area[y][x]&Occ)) return false;
        return true;
    }

    void DrawBlock(unsigned x,unsigned y, int color)
    {
        if(x < (unsigned)Width && y < (unsigned)Height) Area[y][x] = color;
        if(DoDraw) VidMem[y*VidWidth + x + RenderX] = color;
    }

    void DrawPiece(const Piece& piece, int color)
    {
        piece>[&](int x,int y)->bool { this->DrawBlock(x,y,color); return false; };
    }

    bool CollidePiece(const Piece& piece) const
    {
        return piece>[&](int x,int y) { return this->Occupied(x,y); };
    }
    bool CascadeEmpty(int FirstY)
    {
        if(DoDraw)
        {
            ccrBegin(cascadescope);

            // Count full lines
            n_full = list_full = 0;
            for(int y = std::max(0,FirstY); y < Height-1; ++y)
                if(TestFully(y,true))
                {
                    ++n_full;
                    list_full |= 1u << y;
                }
            if(n_full)
            {
                // Clear all full lines in Tengen Tetris style.
                for(animx = 1; animx < Width-1; ++animx)
                {
                    for(timer=Timer; Timer<timer+6; ) ccrReturn(true);
                    auto label = "SINGLEDOUBLETRIPLETETRIS"+(n_full-1)*6;
                    for(int y = FirstY; y < Height-1; ++y)
                        if(list_full & (1u << y))
                            DrawBlock(animx,y,
                                animx-2u < 6u ? label[animx-2] + 0x100 : Empty);
                    if(DoDraw) Sound(10 + animx*n_full*40, 2);
                }
                if(DoDraw) Sound(50, 15);
                // Cascade non-empty lines
                int target = Height-2, y = Height-2;
                for(; y >= 0; --y)
                    if(!(list_full & (1u << y)))
                        DrawRow(target--, [&](unsigned x) { return this->Area[y][x]; });
                // Clear the top lines
                for(auto n=n_full; n-- > 0; )
                    DrawRow(target--, [](unsigned) { return Empty; });
            }
            ccrFinish(false);
        }
        else
        {
            // Cascade non-empty lines
            int target = Height-2, y = Height-2;
            n_full = 0;
            for(int miny = std::max(0,FirstY); y >= miny; --y)
                if(TestFully(y, true))
                {
                    ++n_full;
                    miny = 0;
                }
                else
                {
                    if(target != y)
                        memcpy(&Area[target], &Area[y], sizeof(Area[0][0])*Width);
                    --target;
                }
            // Clear the top lines
            for(auto n=n_full; n-- > 0; )
                DrawRow(target--, [](unsigned) { return Empty; });
            return false;
        }
    }
};

template<bool DoReturns = true>
class TetrisAIengine: TetrisArea<false>
{
public://protected:
    typedef std::pair<int/*score*/, int/*prio*/> scoring;
    struct position
    {
        // 1. Rotate to this position
        int rot;
        // 2. Move to this column
        int x;
        // 3. Drop to ground
        // 4. Rotate to this position
        int rot2;
        // 5. Move maximally to this direction
        int x2;
        // 6. Drop to ground again
    };

    std::vector<position> positions1, positions2;

    struct Recursion
    {
        // Copy of the field before this recursion level
        decltype(Area) bkup;
        // Currently testing position
        unsigned pos_no;
        // Best outcome so far from this recursion level
        scoring  best_score, base_score;
        position best_pos;
        Recursion() : best_pos( {1,5, 1,0} ) { }
    } data[4];
    unsigned ply, ply_limit;

    bool restart;   // Begin a new set of calculations
    bool confident; // false = Reset best-score
    bool resting;   // true  = No calculations to do
    bool doubting;
    struct { ccrVars; } aiscope;

public:
    void AI_Signal(int signal)
    {
        // any = piece moved; 0 = new piece, 2 = at ground
        switch(signal)
        {
            case 0: // new piece
                // do full scan and reset score
                confident = false;
                restart = true;
                resting = false;
                doubting = false;
                break;
            case 1: // piece moved
                // keep scanning (low prio), no resets
                resting = false;
                break;
            case 2: // now at ground
                // do full scan without resetting score
                resting = false;
                restart = true;
                doubting = true;
                break;
        }
    }
    int AI_Run(const decltype(Area)& in_area, const Piece* seq)
    {
        std::memcpy(Area, in_area, sizeof(Area));

        // For AI, we use Pierre Dellacherie's algorithm,
        // but extended for arbitrary ply lookahead.
        enum {
            landingHeightPenalty = -1, erodedPieceCellMetricFactor = 2,
            rowTransitionPenalty = -2, columnTransitionPenalty     = -2,
            buriedHolePenalty    = -8, wellPenalty                 = -2/*,
            occlusionPenalty     = 0*/ };

        ccrBegin(aiscope);

        positions1.clear();
        positions2.clear();
        for(int x=-1; x<Width; ++x)
        for(unsigned rot=0; rot<4; ++rot)
        for(unsigned rot2=0; rot2<4; ++rot2)
        for(int x2=-1; x2<=1; ++x2)
        {
            positions1.push_back( std::move<position> ( {rot,x,rot2,x2} ) );
            if(rot2 == rot && x2 == 0)
                positions2.push_back( std::move<position> ( {rot,x,rot2,x2} ) );
        }

        confident = false;
        doubting  = false;
    Restart:
        restart   = false;
        resting   = false;

        /*std::random_shuffle(positions1.begin(), positions1.end());*/
        std::random_shuffle(positions2.begin(), positions2.end());

        ply = 0;

    Recursion:
        // Take current board as testing platform
        std::memcpy(data[ply].bkup, Area, sizeof(Area));
        if(!confident || ply > 0)
        {
            data[ply].best_score = {-999999,0};
            if(ply == 0)
            {
                //int heap_room = 0;
                //while(TestFully(heap_room,false)) ++heap_room;
              #ifdef EASY
                ply_limit = 2;
              #else
                ply_limit = 3;
                //if(heap_room <= 10) ply_limit = 4;
              #endif
            }
        }
        for(;;)
        {
            data[ply].pos_no = 0;
            do {
                if(DoReturns)
                {
                    ccrReturn(0);
                    if(restart) goto Restart;
                }
                if(ply > 0)
                {
                  // Game might have changed the the board contents.
                    std::memcpy(data[0].bkup, in_area, sizeof(Area));

                    // Now go on and test with the current testing platform
                    std::memcpy(Area, data[ply].bkup, sizeof(Area));
                }

                // Fix the piece in place, cascade rows, and analyze the result.
              { const position& goal = (/*ply==0&&!doubting ? positions1 :*/ positions2)
                                       [ data[ply].pos_no ];
                Piece n = seq[ply];
                n.x = goal.x;
                n.r = goal.rot;
                if(ply) n.y = 0;

                // If we are analyzing a mid-fall maneuver, verify whether
                // the piece can be actually maneuvered into this position.
                //if(ply==0 && n.y >= 0)
                for(Piece q,t=*seq; t.x!=n.x && t.r!=n.r; )
                    if( (t.r == n.r || (q=t, ++t.r, CollidePiece(t)&&(t=q,true)))
                    &&  (t.x <= n.x || (q=t, --t.x, CollidePiece(t)&&(t=q,true)))
                    &&  (t.x >= n.x || (q=t, ++t.x, CollidePiece(t)&&(t=q,true))))
                        goto next_move; // no method of maneuvering.

                // Land the piece if it's not already landed
                do ++n.y; while(!CollidePiece(n)); --n.y;
                if(n.y < 0 || CollidePiece(n)) goto next_move; // cannot place piece?

              /*
                // Rotate to ground-rotation
                if(n.r != goal.rot2 || goal.x2 != 0)
                {
                    while(n.r != goal.rot2)
                    {
                        ++n.r;
                        if(CollidePiece(n)) goto next_move;
                    }
                    if(goal.x2 != 0)
                    {
                        do n.x += goal.x2; while(!CollidePiece(n));
                        n.x -= goal.x2;
                    }

                    do ++n.y; while(!CollidePiece(n)); --n.y;
                    if(n.y < 0 || CollidePiece(n)) goto next_move; // cannot place piece?
                }
              */

                DrawPiece(n, Occ); // place piece

                // Find out the extents of this piece, and how many
                // cells of the piece contribute into full (completed) rows.
                char full[4]={-1,-1,-1,-1};
                int miny=n.y+9, maxy=n.y-9, minx=n.x+9, maxx=n.x-9, num_eroded=0;
                n>[&](int x,int y) -> bool
                    { if(x < minx) minx = x; if(x > maxx) maxx = x;
                      if(y < miny) miny = y; if(y > maxy) maxy = y;
                      if(full[y - n.y] < 0) full[y - n.y] = this->TestFully(y,true);
                      num_eroded += full[y - n.y];
                      return false; };

                CascadeEmpty(n.y);

                // Analyze the board and assign penalties
                int penalties = 0;
                for(int y=0; y<Height-1; ++y)
                    for(int q=1,r, x=1; x<Width; ++x, q=r)
                        if(q != (r = Occupied(x,y)))
                            penalties += rowTransitionPenalty;
                for(int x=1; x<Width-1; ++x)
                    for(int ceil=0/*,heap=0*/,q=0,r, y=0; y<Height; ++y,q=r)
                    {
                        if(q != (r = Occupied(x,y)))
                        {
                            penalties += columnTransitionPenalty;
                            /*if(!r) { penalties += heap * occlusionPenalty; heap = 0; }*/
                        }
                        if(r) { ceil=1; /*++heap;*/ continue; }
                        if(ceil) penalties += buriedHolePenalty;
                        if(Occupied(x-1,y) && Occupied(x+1,y))
                            for(int y2=y; y2<Height-1 && !Occupied(x,y2); ++y2)
                                penalties += wellPenalty;
                    }

                data[ply].base_score = {
                    // score
                    (erodedPieceCellMetricFactor * int(n_full) * num_eroded +
                    penalties +
                    landingHeightPenalty * ((Height-1) * 2 - (miny+maxy))) * 16,
                    // tie-breaker
                    50 * std::abs(Width-2-minx-maxx) +
                    (minx+maxx < Width-2 ? 10 : 0) - n.r
                    - 400*(goal.rot != goal.rot2)
                    - 800*(goal.x2  != 0)
                };
              }
                if(ply+1 < ply_limit)
                    { ++ply; goto Recursion;
                      Unrecursion: --ply; }

                /*fprintf(stdout, "ply %u: [%u]%u,%u gets %d,%d\n",
                    ply,
                    data[ply].pos_no,
                    positions[data[ply].pos_no].x,
                    positions[data[ply].pos_no].rot,
                    data[ply].base_score.first,
                    data[ply].base_score.second);*/

                if(data[ply].best_score < data[ply].base_score)
                {
                    data[ply].best_score  = data[ply].base_score;
                    data[ply].best_pos    = (/*ply==0&&!doubting ? positions1 :*/ positions2) [ data[ply].pos_no ];
                }
            next_move:;
            } while(++data[ply].pos_no < (/*ply==0&&!doubting ? positions1 :*/ positions2).size());

            if(ply > 0)
            {
                int v = data[ply].best_score.first;
                v /= 2;
                //if(ply_limit == 4) v /= 2; else v *= 2;
                data[ply-1].base_score.first += v;// >> (2-ply);
                goto Unrecursion;
                /*
                parent += child / 2 :
                Game 0x7ffff0d94ce0 over with score=91384,lines=92
                Game 0x7ffff0d96a40 over with score=153256,lines=114
                parent += child     :
                Game 0x7fff4a4eb8a0 over with score=83250,lines=86
                Game 0x7fff4a4ed600 over with score=295362,lines=166
                parent += child * 2 :
                Game 0x7fff000a2e00 over with score=182306,lines=131
                Game 0x7fff000a10a0 over with score=383968,lines=193
                parent += child * 4 :
                Game 0x7fff267867b0 over with score=62536,lines=75
                Game 0x7fff26788510 over with score=156352,lines=114
                */
            }
            // all scanned; unless piece placement changes we're as good as we can be.
            confident = true;
            resting   = true;
            doubting  = false;
            while(resting) ccrReturn(0);
            if(restart) goto Restart;
        } // infinite refining loop
        ccrFinish(0);
    }
};

class Tetris: protected TetrisArea<true>
{
public:
    Tetris(unsigned rx) : TetrisArea(rx), seq(),hiscore(0),hudtimer(0) {}
    virtual ~Tetris() { }

protected:
    // These variables should be local to GameLoop(),
    // but because of coroutines, they must be stored
    // in a persistent wrapper instead. Such persistent
    // wrapper is provided by the game object itself.
    Piece seq[4];
    unsigned hiscore, score, lines, combo, pieces;
    unsigned long hudtimer;
    bool escaped, dropping, ticked, kicked, spinned, atground, first;

    struct { ccrVars; } loopscope;
public:
    unsigned incoming;

    int GameLoop()
    {
        Piece &cur = *seq;
        ccrBegin(loopscope);

        // Initialize area
        for(auto y=Height; y-- > 0; )
            for(auto x=Width; x-- > 0; )
                DrawBlock(x,y, (x>0&&x<Width-1&&y<Height-1) ? Empty : Border);

        score = lines = combo = incoming = pieces = 0;
        MakeNext();
        MakeNext();
        first=true;

        for(escaped = false; !escaped; ) // Main loop
        {
            // Generate new piece
            MakeNext();
            /*if(first) // Use if making 4 pieces
            {
                first=false;
                ccrReturn(0);
                MakeNext();
            }*/

            dropping = false;
            atground = false;
        re_collide:
            timer    = Timer;
            AI_Signal(0); // signal changed board configuration

            // Gameover if cannot spawn piece
            if(CollidePiece(cur)) break;

            ccrReturn(0);

            while(!escaped)
            {
                atground = CollidePiece(cur * [](Piece&p){++p.y;});
                if(atground) dropping = false;
                // If we're about to hit the floor, give the AI a chance for sliding.
                AI_Signal(atground ? 2 : 1);

                DrawPiece(cur, cur.color);

                // Wait for input
                for(ticked=false; ; )
                {
                    AI_Run();
                    HUD_Run();
                    ccrReturn(0);

                    if(incoming)
                    {
                        // Receive some lines of garbage from opponent
                        DrawPiece(cur, Empty);
                        for(int threshold=Height-1-incoming, y=0; y<Height-1; ++y)
                        {
                            unsigned mask = 0x1FF7FDFF >> (rand()%10);
                            DrawRow(y, [&](unsigned x) { return
                                y < threshold
                                    ? Area[y+incoming][x]
                                    : (mask & (1<<x)) ? Garbage : Empty; });
                        }
                        // The new rows may push the piece up a bit. Allow that.
                        for(; incoming-- > 0 && CollidePiece(cur); --cur.y) {}
                        incoming = 0;
                        goto re_collide;
                    }

                    ticked = Timer >= timer + std::max(atground?40:10, int( ((17-Level())*8)) );
                    if(ticked || MyKbHit() || dropping) break;
                }

                Piece n = cur;
                if(MyKbHit()) dropping = false;
                switch(ticked ? 's' : (MyKbHit() ? MyGetCh() : (dropping ? 's' : 0)))
                {
                    case 'H'<<8: case 'w': ++n.r; /*Sound(120, 1);*/ break;
                    case 'P'<<8: case 's': ++n.y; break;
                    case 'K'<<8: case 'a': --n.x; /*Sound(120, 1);*/ break;
                    case 'M'<<8: case 'd': ++n.x; /*Sound(120, 1);*/ break;
                    case 'q': case '\033': escaped = true; break;
                    case ' ': dropping = true;
                    default: continue;
                }
                if(n.x != cur.x) kicked = false;

                if(CollidePiece(n))
                {
                    if(n.y == cur.y+1) break; // fix piece if collide against ground
                    // If tried rotating, and was unsuccessful, try wall kicks
                    if(n.r != cur.r && !CollidePiece(n*[](Piece&p){++p.x;})) { kicked = true; ++n.x; } else
                    if(n.r != cur.r && !CollidePiece(n*[](Piece&p){--p.x;})) { kicked = true; --n.x; } else
                    continue; // no move
                }
                DrawPiece(cur, Empty);
                if(n.y > cur.y) timer = Timer; // Reset autodrop timer
                cur = n;
            }

            // If the piece cannot be moved sideways or up from its final position,
            // determine that it must have been spin-fixed into its place.
            // It is a bonus-worthy accomplishment if it ends up clearing lines.
            spinned = CollidePiece(cur*[](Piece&p){--p.y;})
                   && CollidePiece(cur*[](Piece&p){--p.x;})
                   && CollidePiece(cur*[](Piece&p){++p.x;});
            DrawPiece(cur, cur.color|Occ);
            Sound(50, 30);

            while(CascadeEmpty(cur.y)) ccrReturn(0);
            if(n_full > 1) ccrReturn(n_full-1); // Send these rows to opponent

            pieces += 1;
            lines += n_full;
            int multiplier = Level(), clears = ((0x85310 >> (n_full*4)) & 0xF);
            int bonus = (clears * 100 + (cur.y*50/Height));
            int extra = 0;
            if(spinned) extra = ((clears+1) * ((kicked ? 3 : 4)/2) - clears) * 100;
            combo = n_full ? combo + n_full : 0;
            int comboscore = combo > n_full ? combo*50*multiplier : 0;
            bonus *= multiplier;
            extra *= multiplier;
            score += bonus + extra + comboscore;
            HUD_Add(bonus, extra, comboscore);
            //if(n_full) std::fprintf(stdout, "Game %p += %u lines -> score=%u,lines=%u\n", this, n_full, score,lines);
        }
        std::fprintf(stdout, "Game %p over with score=%u,lines=%u\n", this, score,lines);
        over_lines += lines;

        if(score > hiscore) hiscore = score;
        HudPrint(7, 4, "", "%-7u", hiscore);

        ccrFinish(-1);
    }

protected:
    int Level() const { return 1 + lines/10; }

    void MakeNext()
    {
        const int which = 2;
        static const Piece b[] =
        {
            { { { 0x4444,0x00F0,0x4444,0x00F0 } }, 0,0, 0xBDB,0 }, // I
            { { { 0x008E,0x0C88,0x00E2,0x0226 } }, 0,0, 0x1DB,0 }, // J
            { { { 0x002E,0x088C,0x00E8,0x0622 } }, 0,0, 0x6DB,0 }, // L
            { { { 0x0066,0x0066,0x0066,0x0066 } }, 0,0, 0xEDB,0 }, // O 
            { { { 0x08C4,0x006C,0x08C4,0x006C } }, 0,0, 0xADB,0 }, // S
            { { { 0x00E4,0x04C4,0x04E0,0x0464 } }, 0,0, 0x5DB,0 }, // T
            { { { 0x0264,0x00C6,0x0264,0x00C6 } }, 0,0, 0x4DB,0 }, // Z
            // Add some pentaminos to create a challenge worth the AI:
            { { { 0x0272,0x0272,0x0272,0x0272 } }, 0,0, 0x2DB,0 }, // +
            { { { 0x0E44,0x02E2,0x044E,0x08E8 } }, 0,0, 0x9DB,0 }, // T5
            { { { 0x00AE,0x0C8C,0x00EA,0x0626 } }, 0,0, 0x3DB,0 }, // C
            { { { 0x0664,0x0E60,0x2660,0x0670 } }, 0,0, 0x7DB,0 }, // P
            { { { 0x0662,0x0760,0x4660,0x06E0 } }, 0,0, 0xFDB,0 }, // Q
            { { { 0x4444,0x00F0,0x4444,0x00F0 } }, 0,0, 0xBDB,0 }, // I
            { { { 0x0472,0x0362,0x0271,0x0236 } }, 0,0, 0xDDB,0 }, // R
            { { { 0x0172,0x0263,0x0274,0x0632 } }, 0,0, 0x8DB,0 }, // F
            // I is included twice, otherwise it's just a bit too unlikely.
        };
        int c = Empty;
        auto fx = [&seq]() { seq[0].x=4;       seq[0].y=-1;
                             seq[1].x=Width;   seq[1].y=Height-4;
                             seq[2].x=Width+4; seq[2].y=Height-4; };
        auto db = [&](int x,int y) { return VidMem[y*VidWidth+x+RenderX]=c,false; };
        fx();
        seq[1]>db; seq[0] = seq[1];
        seq[2]>db; seq[1] = seq[2];
                   seq[2] = seq[3];
        fx();
        const unsigned npieces = sizeof(b) / sizeof(*b);
        unsigned rnd = (std::rand()/double(RAND_MAX)) * (4*npieces);
    #ifndef EASY
        if(pieces > 3)
        {
            int heap_room = 0;
            while(TestFully(heap_room,false)) ++heap_room;
            bool cheat_good = heap_room <= (7 + lines/20);
            bool cheat_evil = heap_room >= 18;
            if(heap_room >= 16 && (pieces % 5) == 0) cheat_good = false;
            if( (pieces % 11) == 0) cheat_good = true;
            if(cheat_good) cheat_evil = false;
            if(cheat_good || cheat_evil)
            {
                // Don't tell anyone, but in desperate situations,
                // we let AI judge what's best for upcoming pieces,
                // in order to prolong the game!
                // So, the AI cheats, but it is an equal benefactor.
                // EXCEPTION: if there is an abundance of space,
                // bring out the _worst_ possible pieces!
                TetrisAIengine<>::scoring best_score;
                unsigned best_choice = ~0u;
                for(unsigned test=0; test<npieces; ++test)
                {
                    unsigned choice = (test + rnd) % npieces;
                    if(choice == 0) continue; // Ignore the duplicate I (saves time)
                    if(cheat_evil && choice == 7) continue; // Don't give +, it's too evil

                    seq[which]   = b[ choice ];
                    seq[which].r = 0;

                    TetrisAIengine<false> chooser;
                    chooser.AI_Signal(0);
                    chooser.AI_Run(Area, seq);
                    auto& s = chooser.data[0].best_score;
                    if(best_choice == ~0u
                    || (cheat_evil ? s < best_score : s > best_score)
                      )
                        { best_score = s; best_choice = choice; }
                }
                rnd = best_choice * 4 + (rnd % 4);
            }
        }
    #endif
        seq[which] = b[rnd / 4];
        seq[which].r = rnd % 4;
        fx();
        c=seq[1].color;               seq[1]>db;
        c=(seq[2].color&0xF00) + 176; seq[2]>db;
        HudPrint(0,  Height-5, "NEXT:","",0);
        HudPrint(12, 3, "SCORE:", "%-7u", score);
        HudPrint(12, 6, "LINES:", "%-5u", lines);
        HudPrint(12, 9, "LEVEL:", "%-3u", Level());
    }

    void HudPrint(int c, int y,const char*a,const char*b,int v=0) const
    {
        textattr(15); gotoxy(RenderX+Width+2,y);   cprintf(a);
        textattr(c);  gotoxy(RenderX+Width+2,y+1); cprintf(b,v);
    }

    void HUD_Run()
    {
        if(!hudtimer || Timer < hudtimer) return;
        HUD_Add(0,0,0);
        hudtimer = 0;
    }
    void HUD_Add(int bonus, int extra, int combo)
    {
        hudtimer = Timer + 180;
        auto b = "      ";
        HudPrint(10, 11,bonus?b      :b,bonus?"%+-6d":b, bonus);
        HudPrint(10, 14,combo?"Combo":b,combo?"%+-6d":b, combo);
        HudPrint(13, 16,extra?"Skill":b,extra?"%+-6d":b, extra);
    }

    virtual void AI_Signal(int) { }
    virtual int AI_Run() { return 0; }
    virtual int MyKbHit() = 0;
    virtual int MyGetCh() = 0;
public:
    // Return -1 = don't want sleep, +1 = want sleep, 0 = don't mind
    virtual char DelayOpinion() const { return 0; }//dropping ? -1 : 0; }
};

class TetrisHuman: public Tetris
{
public:
    TetrisHuman(unsigned rx) : Tetris(rx) { }
protected:
    virtual int MyKbHit() { return kbhit(); }
    virtual int MyGetCh() { int c; return (c = getch()) ? c : (getch() << 8); }
};

class TetrisAI: public Tetris, TetrisAIengine<true>
{
protected:
    virtual void AI_Signal(int s) { TetrisAIengine::AI_Signal(s); }
    virtual int AI_Run()
    {
      #ifdef __DJGPP__
        char buf1[16]/*, buf2[16], buf3[16]*/;
        sprintf(buf1, "COM%u%c%c%c%c", ply_limit,
            resting   ? 'r' : '_',
            confident ? 'C' : '_',
            doubting  ? 'D' : '_',
            atground  ? 'G' : '_'
        );
        ScreenPutString(buf1, 0x70, Tetris::RenderX+Width, 0);
        /*
        sprintf(buf2, "%-8d",   data[0].best_score.first);
        sprintf(buf3, "%2d,%d", data[0].best_pos.x,
                                data[0].best_pos.rot);
        ScreenPutString(buf2, 0x08, Tetris::RenderX+Width, 1);
        ScreenPutString(buf3, 0x08, Tetris::RenderX+Width, 4);
        */
      #endif
        return TetrisAIengine::AI_Run(Tetris::Area, seq);
    }
    virtual int MyKbHit()
    {
        return PendingAIinput ? 1 : (PendingAIinput = GenerateAIinput()) != 0;
    }
    virtual int MyGetCh()
    {
        int r = PendingAIinput;
        return r ? (PendingAIinput=0), r : GenerateAIinput();
    }
    char d(char c, int maxdelay)
    {
        if(Timer >= intimer+maxdelay) { intimer=Timer; return c; }
        return 0;
    }
    int GenerateAIinput()
    {
        /*if(TetrisAIengine::atground)
        {
            if(seq->r != data[0].best_pos.rot2) return d('w',1);
            if(data[0].best_pos.x2 < 0) return d('a', 1);
            if(data[0].best_pos.x2 > 0) return d('d', 1);
        }
        else*/
        {
            if(seq->r != data[0].best_pos.rot) return d('w',1);
            if(seq->y >= 0 && seq->x > data[0].best_pos.x) return d('a', 2);
            if(seq->y >= 0 && seq->x < data[0].best_pos.x) return d('d', 2);
        }
        if(doubting || !confident) return 0;
        return d('s',3);
    }

    int PendingAIinput, delay;
    unsigned long intimer;
public:
    virtual char DelayOpinion() const
        { if(doubting || restart || !confident) return -1;
          return Tetris::DelayOpinion(); }

    TetrisAI(unsigned rx) : Tetris(rx), PendingAIinput(0), delay(0), intimer(0) {}
};

int main()
{
    const unsigned NewTimerFreq = 173;//519;//173;//121
    unsigned TimerPeriod = 0x1234DDul / NewTimerFreq;

    textmode(C40);

    outportb(0x43, 0x34);
    outportb(0x40, TimerPeriod & 0xFF);
    outportb(0x40, TimerPeriod >>   8);

    //TetrisHuman player(0);
    TetrisAI player(0);
    TetrisAI computer(20);
    //textattr(0x78); gotoxy(13,1); cprintf("1UP ");
    //textattr(0x78); gotoxy(13,1); cprintf("COM@3 ");

    unsigned overs = 0;
    while(overs < 40)
    {
        int a,b;
    //#pragma omp parallel sections shared(a,b) num_threads(2)
    {
      //#pragma omp section
      {
        a = player.GameLoop();
      }
      //#pragma omp section
      {
        b = computer.GameLoop();
      }
    }
        if(a < 0) ++overs;
        if(b < 0) ++overs;
        if(a < 0 || b < 0) continue;
      #ifndef EASY
        player.incoming   += b;
        computer.incoming += a;
      #endif

        if(computer.DelayOpinion() >= 0 && player.DelayOpinion() >= 0)
        {
            //__dpmi_yield(); // Low-power delay, if nothing to do
            #ifdef __DJGPP__
            __asm__ volatile("hlt");
            #else
            timerdelay=0;
            #endif
        }
        UpdateSound();
    }
    std::printf("Average lines before gameover: %u\n", over_lines / overs);

    // Gameover
    textmode(C80);
    std::puts("GAME OVER");
}
