<?php

/* Tetromino / pentomino / hexamino generator program
 * for piece puzzle games,
 * created and (C) by Joel Yliluoma in 2012
 * http://iki.fi/bisqwit/
 */

define('MAPDIM', 8); // Dimensions of the map in which we render

define('MAXIMUM_DIMENSION', 4); // Dimensions of the piece within the game
define('MINIMUM_SIZE',      4); // Minimum number of blocks in the piece
define('MAXIMUM_SIZE',      5); // Maximum number of blocks in the piece

define('MAXTRIES', 1000);

define('BLANK',  '.');
define('PIECE',  '#');

// Binary to hexadecimal conversion array...
$bin = Array('....'=>0, '...#'=>1, '..#.'=>2, '..##'=>3,
             '.#..'=>4, '.#.#'=>5, '.##.'=>6, '.###'=>7,
             '#...'=>8, '#..#'=>9, '#.#.'=>10,'#.##'=>11,
             '##..'=>12,'##.#'=>13,'###.'=>14,'####'=>15);

function shuffle_assoc( $array )
{
   $keys = array_keys( $array );
   shuffle( $keys );
   return array_merge( array_flip( $keys ) , $array );
} 

$done = Array();

$size  = MINIMUM_SIZE;
$tries = 0;

for(;;)
{
  if(++$tries > MAXTRIES)
  {
    $tries = 0;
    ++$size;
    if($size > MAXIMUM_SIZE) break;
  }

  $blank = str_pad('', MAPDIM*MAPDIM, BLANK);
  $remain = Array();
  
  /* Generate piece */
  
  $s = $blank;
  $parts = Array();
  
  $pset = function($x,$y) use(&$s, &$remain, &$parts)
  {
    $minx = MAPDIM; $miny = MAPDIM;
    $maxx = 0; $maxy = 0;
    if($x < $minx) $minx = $x;
    if($y < $miny) $miny = $y;
    if($x > $maxx) $maxx = $x;
    if($y > $maxy) $maxy = $y;
    $pos = $y*MAPDIM+$x;
    $s[$pos] = PIECE;
    unset($remain[$pos]);
    if($y > 0 && $s[$pos-MAPDIM]==BLANK) $remain[$pos-MAPDIM] = $pos-MAPDIM;
    if($y < (MAPDIM-1) && $s[$pos+MAPDIM]==BLANK) $remain[$pos+MAPDIM] = $pos+MAPDIM;
    if($x > 0 && $s[$pos-1]==BLANK) $remain[$pos-1] = $pos-1;
    if($x < (MAPDIM-1) && $s[$pos+1]==BLANK) $remain[$pos+1] = $pos+1;
    
    $parts[] = Array(0.5 + $x, 0.5 + $y);
  };
  
  $pset(MAPDIM >> 1, MAPDIM >> 1);

  for($n=1; $n<$size; ++$n)
  {
    $r = $remain;
    $remain = Array();
    shuffle($r);
    foreach($r as $loc) $remain[$loc] = $loc;
    $pset($loc % MAPDIM, (int)($loc / MAPDIM));
  }

  /* Figure out center */
  if(0)
  {
    $centerx = 0;
    $centery = 0;
    foreach($parts as $part)
    {
      $centerx += $part[0];
      $centery += $part[1];
    }
    $centerx /= count($parts);
    $centery /= count($parts);
  }
  else
  {
    $minx = MAPDIM; $miny = MAPDIM;
    $maxx = 0; $maxy = 0;
    foreach($parts as $part)
    {
      if($part[0] < $minx) $minx = $part[0];
      if($part[1] < $miny) $miny = $part[1];
      if($part[0] > $maxx) $maxx = $part[0];
      if($part[1] > $maxy) $maxy = $part[1];
    }
    $centerx = ($minx + $maxx) * 0.5;
    $centery = ($miny + $maxy) * 0.5;
  }

  /* Generate rotated versions of the piece */
  $forms = Array($s);
  for($rotation=0; $rotation<3; ++$rotation)
  {
    $new = $blank;
    $s = (int)(1.2*sin(($rotation+1)*3.141592653/2));
    $c = (int)(1.2*cos(($rotation+1)*3.141592653/2));
    foreach($parts as $part)
    {
      $x = $centerx + ($part[0]-$centerx) * $c - ($part[1]-$centery) * $s;
      $y = $centery + ($part[1]-$centery) * $c + ($part[0]-$centerx) * $s;
      $pos = (int)($y) * MAPDIM + (int)($x);
      if($pos >= 0 && $pos < MAPDIM*MAPDIM)
        $new[$pos] = PIECE;
    }
    $forms[] = $new;
  }
  /* Figure out containing box */
  $minx = MAPDIM; $miny = MAPDIM;
  $maxx = 0; $maxy = 0;
  foreach($forms as $s)
  {
    for($p=$y=0; $y<MAPDIM; ++$y)
      for($x=0; $x<MAPDIM; ++$x, ++$p)
        if($s[$p] != BLANK)
        {
          if($x < $minx) $minx = $x;
          if($y < $miny) $miny = $y;
          if($x > $maxx) $maxx = $x;
          if($y > $maxy) $maxy = $y;
        }
  }
  if($maxx-$minx > MAXIMUM_DIMENSION
  || $maxy-$miny > MAXIMUM_DIMENSION)
  {
    continue;
  }

  /* Adjust all pieces to the lower-right corner */
  foreach($forms as &$s)
  {
    $ox = (MAPDIM-1)-$maxx;
    $oy = (MAPDIM-1)-$maxy;
    $o = $oy*MAPDIM+$ox;
    $new = $blank;
    for($p=$y=0; $y<MAPDIM; ++$y)
      for($x=0; $x<MAPDIM; ++$x, ++$p)
        if($s[$p] == PIECE)
          $new[$p+$o] = PIECE;
    $s = $new;
  }
  unset($s);

  foreach($forms as $s)
    if(isset($done[$s]))
      continue 2; /* Ignore this piece, already synthesized */

  /* Add the piece to done-list */
  foreach($forms as $s)
    $done[$s] = true;

  /* Reformat the piece in the format expected by the game engine */
  $new = Array('', '', '', '');
  $o = MAPDIM-MAXIMUM_DIMENSION;
  for($n=0; $n<4; ++$n)
    for($y=0; $y<MAXIMUM_DIMENSION; ++$y)
      $new[$n] .= substr($forms[$n], ($y+$o)*MAPDIM + $o, MAXIMUM_DIMENSION);

  if(true)
  {
    /* Debug render */
    for($y=0; $y<MAXIMUM_DIMENSION; ++$y)
    {
      print "/* ";
      for($n=0; $n<4; ++$n)
        print substr($new[$n], $y*MAXIMUM_DIMENSION, MAXIMUM_DIMENSION).' ';
      print " */\n";
    }
    print "//\n";
  }
  
  print "{ { { ";
  for($n=0; $n<4; ++$n)
  {
    $s = $new[$n];
    while(strlen($s) % 4) $s = ".$s";
    print "0x";
    for($p=0; $p<strlen($s)/4; ++$p)
      printf('%X', $bin[substr($s,$p*4,4)]);
    print "ULL";
    if($n+1<4)print ",";
  }
  $color = 1 + ((crc32(serialize($forms))&0xFFFFFFF) % 15);
  printf(" } }, 0,0, 0x%XDB,0 }, //\n", $color);
  
  $tries = 0;
}
